#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Copyright 2012 Liricus SRL <info@liricus.com.ar>

# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Fra


#===============================================================================
# DOCS
#===============================================================================

"""Defines database models and api for facebook integration

"""


#===============================================================================
# IMPORTS
#===============================================================================

import hashlib
import time
import os
import random

from poster.encode import multipart_encode
from poster.streaminghttp import register_openers
import urllib2

import urllib
import urllib2

import pyfb

import flask

from flask_peewee import db

from socialframe_server import settings
from socialframe_server import core
from socialframe_server import models
from socialframe_server.ext import multipart


#===============================================================================
# CONFS
#===============================================================================

# Register the streaming http handlers with urllib2
register_openers()


#===============================================================================
# CONSTANTS
#===============================================================================

AUTH_SUCCESS_URL = "http://{host}:{port}/facebook/auth_success".format(host=settings.HOST,
                                                                        port=settings.PORT)


#===============================================================================
# TABLES
#===============================================================================

@models.autocreatetable
class FBUser(models.SocialFrameModel):

    user = db.ForeignKeyField(models.User, unique=True)
    auth_token = db.CharField(null=True)
    redirect = db.TextField(null=True)
    facebook_id = db.CharField(null=True)
    sf_token = db.CharField(null=True)

    @classmethod
    def get_loggedin_user(cls):
        user = core.APP_AUTH.get_logged_in_user()
        return FBUser.get_or_create(user=user)

    def __unicode__(self):
        return self.user.email


@models.autocreatetable
class FBAlbum2BubbleSync(models.SocialFrameModel):

    fuser = db.ForeignKeyField(FBUser)
    bubble = db.ForeignKeyField(models.Bubble)
    album_id = db.CharField()
    album_name = db.CharField()
    size = db.CharField()

    def to_dict(self):
        return {"bubble": self.bubble.name,
                 "album_name": self.album_name,
                 "size": self.size,
                 "album_id": self.album_id}

    def save(self, force_insert=False):
        if self.size not in settings.IMAGE_SIZES:
            raise ValueError("invalid image size: " +  self.size)
        super(FBAlbum2BubbleSync, self).save(force_insert)

    def __unicode__(self):
        return "{} -> {}".format(self.bubble.name, self.album_id)


@models.autocreatetable
class FBSyncImage(models.SocialFrameModel):

    image = db.ForeignKeyField(models.Image)
    sync = db.ForeignKeyField(FBAlbum2BubbleSync,
                              related_name="sync_images")


#===============================================================================
# API
#===============================================================================

@core.api("/facebook/authenticate",
            loggedin=True, cached=60, transaction=True, json=False,
            methods=["POST"])
def authenticate():
    """Authenticate the user in facebook

    **Logged in:** True

    **WARNING** Not Json

    **Method:** POST

    **Params:**
        :redirect: The landing page after authentification.

    **Return:**

        ::

            {"error": false,
             "action": "authenticate",
             "code": 200}

    """
    fuser = FBUser.get_loggedin_user()
    fuser.redirect = urllib.unquote_plus(flask.request.form["redirect"])
    fuser.sf_token = hashlib.sha256(fuser.user.email +
                                    str(time.time()) +
                                    str(random.random)).hexdigest()
    facebook = pyfb.Pyfb(settings.FACEBOOK_APP_ID)
    fuser.save()
    auth_success = "{}/{}/{}".format(AUTH_SUCCESS_URL,
                                      fuser.user.email, fuser.sf_token)
    return flask.redirect(facebook.get_auth_code_url(redirect_uri=auth_success))


@core.api("/facebook/auth_success/<email:email>/<sf_token>",
            cached=60, transaction=True, json=False)
def auth_success(email, sf_token):
    """This method is called from facebook... don't use it"""
    fuser = FBUser.get(FBUser.sf_token == sf_token)
    code = flask.request.args['code']
    facebook = pyfb.Pyfb(settings.FACEBOOK_APP_ID)
    auth_success = "{}/{}/{}".format(AUTH_SUCCESS_URL,
                                      fuser.user.email, fuser.sf_token)
    fuser.auth_token = facebook.get_access_token(settings.FACEBOOK_SECRET_KEY,
                                                 code,
                                                 redirect_uri=auth_success)
    fuser.sf_token = None
    fuser.facebook_id = facebook.get_myself().id
    fuser.save()
    return flask.redirect(fuser.redirect)


@core.api("/facebook/is_authenticated", loggedin=True, cached=60)
def is_authenticated():
    """Returns true if the user is authentificated in facebook

    **Logged in:** True

    **Method:** GET

    **Return**

        ::

            {"error": false,
             "action": "is_authenticated",
             "is_authenticated": true|false,
             "code": 200}

    """
    fuser = FBUser.get_loggedin_user()
    ia = False
    if fuser.auth_token is not None:
        fb = pyfb.Pyfb(settings.FACEBOOK_APP_ID)
        fb.set_access_token(fuser.auth_token)
        try:
            fb.get_myself()
        except pyfb.PyfbException:
            ia = False
        else:
            ia = True
    return {"is_authenticated": ia}


@core.api("/facebook/sync_bubble",
            loggedin=True, cached=120, transaction=True, methods=["POST"])
def sync_bubble():
    fuser = FBUser.get_loggedin_user()
    fb = pyfb.Pyfb(settings.FACEBOOK_APP_ID)
    fb.set_access_token(fuser.auth_token)

    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.user == fuser.user,
                               models.Bubble.name == bubble_name)

    sync = FBAlbum2BubbleSync.get(FBAlbum2BubbleSync.bubble == bubble)
    imgs = [s.image.id for s in sync.sync_images]

    query = models.Image.select()
    query = query.where(models.Image.bubble==bubble,
                        models.Image.status==models.Image.STATUS_APPROVED,
                        ~(models.Image.id << imgs))
    url = "https://graph.facebook.com/{}/photos".format(sync.album_id)
    for img in query:
        imgsize = img.get_size(sync.size)
        params = {"photo": open(imgsize.path, "rb"),
                  "message": img.desc,
                  "access_token": fuser.auth_token}
        datagen, headers = multipart_encode(params)
        request = urllib2.Request(url, datagen, headers)
        urllib2.urlopen(request).read()
        sync_img = FBSyncImage()
        sync_img.sync = sync
        sync_img.image = img
        sync_img.save()
    return {"syncs": query.count()}


@core.api("/facebook/list_sync", loggedin=True, cached=60)
def list_sync():
    fuser = FBUser.get_loggedin_user()
    query = FBAlbum2BubbleSync.select()
    query = query.where(FBAlbum2BubbleSync.fuser == fuser)
    return {"syncs": [s.to_dict() for s in query]}


@core.api("/facebook/is_sync/<bubble_name>", loggedin=True, cached=60)
def is_sync(bubble_name):
    fuser = FBUser.get_loggedin_user()
    bubble_name = urllib.unquote_plus(bubble_name)
    bubble = models.Bubble.get(models.Bubble.user == fuser.user,
                               models.Bubble.name == bubble_name)
    query = FBAlbum2BubbleSync.select()
    query = query.where(FBAlbum2BubbleSync.bubble==bubble)
    return {"is_sync": bool(query.count())}


@core.api("/facebook/add_sync",
            loggedin=True, cached=60, transaction=True, methods=["POST"])
def add_sync():
    fuser = FBUser.get_loggedin_user()
    fb = pyfb.Pyfb(settings.FACEBOOK_APP_ID)
    fb.set_access_token(fuser.auth_token)

    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.user == fuser.user,
                               models.Bubble.name == bubble_name)

    query = FBAlbum2BubbleSync.select()
    query = query.where(FBAlbum2BubbleSync.bubble == bubble)
    if query.count():
        flask.abort(400)

    # determine the name of the new album on facebook
    base_name = "{}_{}".format(bubble.name, settings.APP_NAME)
    idx = 0
    new_name = base_name
    albums = [a.name for a in fb._client.get_list(fuser.facebook_id, "albums")]
    while new_name in albums:
        new_name = "{}_{}".format(base_name, idx)

    new_sync = FBAlbum2BubbleSync()
    new_sync.fuser = fuser
    new_sync.bubble = bubble
    new_sync.size = urllib.unquote_plus(flask.request.form.get("size", settings.BIG_SIZE))
    new_sync.album_name = new_name
    new_sync.album_id = fb._client.push(fuser.facebook_id, "albums",
                                        message=bubble.desc, name=new_name).id
    new_sync.save()
    return {"new_album": new_name}


@core.api("/facebook/delete_sync",
            loggedin=True, cached=60, transaction=True, methods=["POST"])
def delete_sync():
    fuser = FBUser.get_loggedin_user()
    fb = pyfb.Pyfb(settings.FACEBOOK_APP_ID)
    fb.set_access_token(fuser.auth_token)

    bubble_name = urllib.unquote_plus(flask.request.form["bubble_name"])
    bubble = models.Bubble.get(models.Bubble.user == fuser.user,
                               models.Bubble.name == bubble_name)

    sync = FBAlbum2BubbleSync.get(FBAlbum2BubbleSync.bubble == bubble)
    sync.delete_instance(True)


#===============================================================================
# MAIN
#===============================================================================

if __name__ == "__main__":
    print(__doc__)
